import pandas as pd
import nibabel as nb
from nilearn import plotting as nlp
import nistats as nis
import nistats.reporting  # noqa: F401

from nipype.interfaces.base import (
    SimpleInterface, BaseInterfaceInputSpec, TraitedSpec,
    File, traits
    )
from nipype.utils.filemanip import fname_presuffix, split_filename

from ..viz import plot_and_save, plot_corr_matrix, plot_contrast_matrix


class VisualizationInputSpec(BaseInterfaceInputSpec):
    data = File(mandatory=True, desc='Data file to visualize')
    image_type = traits.Enum('svg', 'png', mandatory=True)


class VisualizationOutputSpec(TraitedSpec):
    figure = File(desc='Visualization')


class Visualization(SimpleInterface):
    input_spec = VisualizationInputSpec
    output_spec = VisualizationOutputSpec

    def _run_interface(self, runtime):
        import matplotlib
        matplotlib.use('Agg')
        import seaborn as sns
        from matplotlib import pyplot as plt
        sns.set_style('white')
        plt.rcParams['svg.fonttype'] = 'none'
        plt.rcParams['image.interpolation'] = 'nearest'

        data = self._load_data(self.inputs.data)
        out_name = fname_presuffix(self.inputs.data,
                                   suffix='.' + self.inputs.image_type,
                                   newpath=runtime.cwd,
                                   use_ext=False)
        self._visualize(data, out_name)
        self._results['figure'] = out_name
        return runtime

    def _load_data(self, fname):
        _, _, ext = split_filename(fname)
        if ext == '.tsv':
            return pd.read_table(fname, index_col=0)
        elif ext in ('.nii', '.nii.gz', '.gii'):
            return nb.load(fname)
        raise ValueError("Unknown file type!")


class DesignPlot(Visualization):
    def _visualize(self, data, out_name):
        from matplotlib import pyplot as plt
        plt.set_cmap('viridis')
        plot_and_save(out_name, nis.reporting.plot_design_matrix, data)


class DesignCorrelationPlotInputSpec(VisualizationInputSpec):
    explanatory_variables = traits.Range(
        low=0, desc='Number of explanatory variables')


class DesignCorrelationPlot(Visualization):
    input_spec = DesignCorrelationPlotInputSpec

    def _visualize(self, data, out_name):
        plot_and_save(out_name, plot_corr_matrix,
                      data.drop(columns='constant').corr(),
                      self.inputs.explanatory_variables)


class ContrastMatrixPlotInputSpec(VisualizationInputSpec):
    orientation = traits.Enum('horizontal', 'vertical', usedefault=True,
                              desc='Display orientation of contrast matrix')


class ContrastMatrixPlot(Visualization):
    input_spec = ContrastMatrixPlotInputSpec

    def _visualize(self, data, out_name):
        if 'constant' in data.index:
            data = data.drop(index='constant')
        plot_and_save(out_name, plot_contrast_matrix, data,
                      ornt=self.inputs.orientation)


class GlassBrainPlot(Visualization):
    def _visualize(self, data, out_name):
        nlp.plot_glass_brain(data, colorbar=True, plot_abs=False,
                             display_mode='lyrz', axes=None,
                             output_file=out_name)



from os import path as op
from pathlib import Path
import jinja2
import pkg_resources as pkgr
import bids

from ..utils import snake_to_camel

PATH_PATTERNS = [
    '[sub-{subject}/][ses-{session}/][sub-{subject}_][ses-{session}_]'
    'model-{model}.html'
]


def deroot(val, root):
    if isinstance(val, str):
        if val.startswith(root):
            idx = len(root)
            if val[idx] == '/':
                idx += 1
            val = val[idx:]
    elif isinstance(val, list):
        val = [deroot(elem, root) for elem in val]
    elif isinstance(val, dict):
        val = {key: deroot(value, root) for key, value in val.items()}

    return val


def parse_directory(deriv_dir, work_dir, analysis):
    fl_layout = bids.BIDSLayout(
        (deriv_dir, ['bids', 'derivatives',
                     pkgr.resource_filename('fitlins', 'data/fitlins.json')]))
    wd_layout = bids.BIDSLayout(str(Path(work_dir) / 'reportlets' / 'fitlins'))
    contrast_svgs = fl_layout.get(extensions='.svg', type='contrasts')

    analyses = []
    for contrast_svg in contrast_svgs:
        ents = fl_layout.parse_file_entities(contrast_svg.filename)
        ents.pop('type')
        ents.setdefault('subject', None)
        correlation_matrix = fl_layout.get(extensions='.svg', type='corr',
                                           **ents)
        design_matrix = fl_layout.get(extensions='.svg', type='design', **ents)
        job_desc = {
            'ents': {k: v for k, v in ents.items() if v is not None},
            'dataset': analysis.layout.root,
            'model_name': analysis.model['name'],
            'contrasts_svg': contrast_svg.filename,
            }
        if ents.get('subject'):
            job_desc['subject_id'] = ents.get('subject')
        if correlation_matrix:
            job_desc['correlation_matrix_svg'] = correlation_matrix[0].filename
        if design_matrix:
            job_desc['design_matrix_svg'] = design_matrix[0].filename

        snippet = wd_layout.get(extensions='.html', type='snippet', **ents)
        if snippet:
            with open(snippet[0].filename) as fobj:
                job_desc['warning'] = fobj.read()

        contrasts = fl_layout.get(extensions='.png', type='ortho', **ents)
        # TODO: Split contrasts from estimates
        job_desc['contrasts'] = [{'image_file': c.filename,
                                  'name':
                                      fl_layout.parse_file_entities(
                                          c.filename)['contrast']}
                                 for c in contrasts]
        analyses.append(job_desc)

    return analyses


def write_report(level, report_dicts, run_context, deriv_dir):
    fl_layout = bids.BIDSLayout(
        (deriv_dir, ['bids', 'derivatives',
                     pkgr.resource_filename('fitlins', 'data/fitlins.json')]))

    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(
            searchpath=pkgr.resource_filename('fitlins', '/')))

    tpl = env.get_template('data/report.tpl')

    for context in report_dicts:
        ents = context['ents'].copy()
        ents['model'] = snake_to_camel(context['model_name'])
        target_file = op.join(deriv_dir, fl_layout.build_path(ents, PATH_PATTERNS))
        html = tpl.render(deroot({'level': level, **context, **run_context},
                                 op.dirname(target_file)))
        with open(target_file, 'w') as fobj:
            fobj.write(html)



from pathlib import Path
from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu
from ..interfaces.bids import (
    ModelSpecLoader, LoadBIDSModel, BIDSSelect, BIDSDataSink)
from ..interfaces.nistats import FirstLevelModel, SecondLevelModel
from ..interfaces.visualizations import (
    DesignPlot, DesignCorrelationPlot, ContrastMatrixPlot, GlassBrainPlot)
from ..interfaces.utils import MergeAll


def init_fitlins_wf(bids_dir, preproc_dir, out_dir, space, exclude_pattern=None,
                    include_pattern=None, model=None, participants=None,
                    base_dir=None, name='fitlins_wf'):
    wf = pe.Workflow(name=name, base_dir=base_dir)

    # Find the appropriate model file(s)
    specs = ModelSpecLoader(bids_dir=bids_dir)
    if model is not None:
        specs.inputs.model = model

    all_models = specs.run().outputs.model_spec
    if not all_models:
        raise RuntimeError("Unable to find or construct models")

    #
    # Load and run the model
    #

    selectors = {'subject': participants} if participants is not None else {}

    loader = pe.Node(
        LoadBIDSModel(bids_dir=bids_dir,
                      preproc_dir=preproc_dir,
                      selectors=selectors),
        name='loader')

    if preproc_dir is not None:
        loader.inputs.preproc_dir = preproc_dir
    if exclude_pattern is not None:
        loader.inputs.exclude_pattern = exclude_pattern
    if include_pattern is not None:
        loader.inputs.include_pattern = include_pattern

    if isinstance(all_models, list):
        loader.iterables = ('model', all_models)
    else:
        loader.inputs.model = all_models

    # Because pybids generates the entire model in one go, we will need
    # various helper nodes to select the correct portions of the model
    select_l1_entities = pe.Node(niu.Select(index=0), name='select_l1_entities')

    # Select preprocessed BOLD series to analyze
    getter = pe.Node(
        BIDSSelect(bids_dir=bids_dir,
                   selectors={'type': 'preproc', 'space': space}),
        name='getter')

    if preproc_dir is not None:
        getter.inputs.preproc_dir = preproc_dir

    select_l1_contrasts = pe.Node(niu.Select(index=0), name='select_l1_contrasts')

    # Run first level model
    l1_model = pe.MapNode(
        FirstLevelModel(),
        iterfield=['session_info', 'contrast_info', 'bold_file', 'mask_file'],
        name='l1_model')

    def join_dict(base_dict, dict_list):
        return [{**base_dict, **iter_dict} for iter_dict in dict_list]

    # Accumulate metadata
    l1_metadata = pe.MapNode(niu.Function(function=join_dict),
                             iterfield=['base_dict', 'dict_list'],
                             name='l1_metadata')

    # Squash the results of MapNodes that may have generated multiple maps
    # into single lists.
    # Do the same with corresponding metadata - interface will complain if shapes mismatch
    collate_first_level = pe.Node(MergeAll(['contrast_maps', 'contrast_metadata']),
                                  name='collate_first_level')

    select_l2_entities = pe.Node(niu.Select(index=1), name='select_l2_entities')
    select_l2_indices = pe.Node(niu.Select(index=1), name='select_l2_indices')
    select_l2_contrasts = pe.Node(niu.Select(index=1), name='select_l2_contrasts')

    # Run second-level model
    # TODO: Iterate over all higher levels
    l2_model = pe.MapNode(
        SecondLevelModel(),
        iterfield=['contrast_info', 'contrast_indices'],
        name='l2_model')

    collate_second_level = pe.Node(MergeAll(['contrast_maps', 'contrast_metadata']),
                                   name='collate_second_level')

    #
    # Plotting
    #

    plot_design = pe.MapNode(
        DesignPlot(image_type='svg'),
        iterfield='data',
        name='plot_design')

    def _get_evs(info):
        import pandas as pd
        events = pd.read_hdf(info['events'], key='events')
        return len(events['condition'].unique())

    # Number of explanatory variables is used to mark off sections of the
    # correlation matrix
    get_evs = pe.MapNode(niu.Function(function=_get_evs), iterfield='info', name='get_evs')

    plot_corr = pe.MapNode(
        DesignCorrelationPlot(image_type='svg'),
        iterfield=['data', 'explanatory_variables'],
        name='plot_corr')

    plot_l1_contrast_matrix = pe.MapNode(
        ContrastMatrixPlot(image_type='svg'),
        iterfield='data',
        name='plot_l1_contrast_matrix')

    plot_l1_contrasts = pe.MapNode(
        GlassBrainPlot(image_type='png'),
        iterfield='data',
        name='plot_l1_contrasts')

    plot_l2_contrast_matrix = pe.MapNode(
        ContrastMatrixPlot(image_type='svg'),
        iterfield='data',
        name='plot_l2_contrast_matrix')

    plot_l2_contrasts = pe.MapNode(
        GlassBrainPlot(image_type='png'),
        iterfield='data',
        name='plot_l2_contrasts')

    #
    # HTML snippets to be included directly in report, not
    # saved as individual derivative files
    #

    reportlet_dir = Path(base_dir) / 'reportlets' / 'fitlins'
    reportlet_dir.mkdir(parents=True, exist_ok=True)

    snippet_pattern = '[sub-{subject}/][ses-{session}/][sub-{subject}_]' \
        '[ses-{session}_]task-{task}_[run-{run}_]snippet.html'
    ds_model_warnings = pe.MapNode(
        BIDSDataSink(base_directory=str(reportlet_dir),
                     path_patterns=snippet_pattern),
        iterfield=['entities', 'in_file'],
        run_without_submitting=True,
        name='ds_model_warning')

    #
    # Derivatives
    #

    # NIfTIs
    contrast_pattern = '[sub-{subject}/][ses-{session}/]' \
        '[sub-{subject}_][ses-{session}_]task-{task}[_acq-{acquisition}]' \
        '[_rec-{reconstruction}][_run-{run}][_echo-{echo}]_bold' \
        '[_space-{space}]_contrast-{contrast}_{type<effect|stat>}.nii.gz'
    ds_l1_contrast_maps = pe.Node(
        BIDSDataSink(base_directory=out_dir,
                     path_patterns=contrast_pattern),
        run_without_submitting=True,
        name='ds_l1_contrast_maps')

    ds_l2_contrast_maps = pe.Node(
        BIDSDataSink(base_directory=out_dir,
                     path_patterns=contrast_pattern),
        run_without_submitting=True,
        name='ds_l2_contrast_maps')

    # Images
    image_pattern = '[sub-{subject}/][ses-{session}/]' \
        '[sub-{subject}_][ses-{session}_]task-{task}[_acq-{acquisition}]' \
        '[_rec-{reconstruction}][_run-{run}][_echo-{echo}]_bold_' \
        '{type<design|corr|contrasts>}.svg'
    ds_design = pe.MapNode(
        BIDSDataSink(base_directory=out_dir, fixed_entities={'type': 'design'},
                     path_patterns=image_pattern),
        iterfield=['entities', 'in_file'],
        run_without_submitting=True,
        name='ds_design')

    ds_corr = pe.MapNode(
        BIDSDataSink(base_directory=out_dir, fixed_entities={'type': 'corr'},
                     path_patterns=image_pattern),
        iterfield=['entities', 'in_file'],
        run_without_submitting=True,
        name='ds_corr')

    ds_l1_contrasts = pe.MapNode(
        BIDSDataSink(base_directory=out_dir,
                     fixed_entities={'type': 'contrasts'},
                     path_patterns=image_pattern),
        iterfield=['entities', 'in_file'],
        run_without_submitting=True,
        name='ds_l1_contrasts')

    ds_l2_contrasts = pe.MapNode(
        BIDSDataSink(base_directory=out_dir,
                     fixed_entities={'type': 'contrasts'},
                     path_patterns=image_pattern),
        iterfield=['entities', 'in_file'],
        run_without_submitting=True,
        name='ds_l2_contrasts')

    contrast_plot_pattern = '[sub-{subject}/][ses-{session}/]' \
        '[sub-{subject}_][ses-{session}_]task-{task}[_acq-{acquisition}]' \
        '[_rec-{reconstruction}][_run-{run}][_echo-{echo}]_bold' \
        '[_space-{space}]_contrast-{contrast}_ortho.png'
    ds_l1_contrast_plots = pe.Node(
        BIDSDataSink(base_directory=out_dir,
                     path_patterns=contrast_plot_pattern),
        run_without_submitting=True,
        name='ds_l1_contrast_plots')

    ds_l2_contrast_plots = pe.Node(
        BIDSDataSink(base_directory=out_dir,
                     path_patterns=contrast_plot_pattern),
        run_without_submitting=True,
        name='ds_l2_contrast_plots')

    #
    # Connections
    #

    wf.connect([
        #
        # Modeling
        #
        (loader, select_l1_entities, [('entities', 'inlist')]),

        (select_l1_entities, getter,  [('out', 'entities')]),

        (loader, select_l1_contrasts, [('contrast_info', 'inlist')]),

        (loader, l1_model, [('session_info', 'session_info')]),
        (getter, l1_model, [('bold_files', 'bold_file'),
                            ('mask_files', 'mask_file')]),
        (select_l1_contrasts, l1_model,  [('out', 'contrast_info')]),

        (getter, l1_metadata, [('entities', 'base_dict')]),
        (l1_model, l1_metadata, [('contrast_metadata', 'dict_list')]),

        (l1_model, collate_first_level, [('contrast_maps', 'contrast_maps')]),
        (l1_metadata, collate_first_level, [('out', 'contrast_metadata')]),

        (loader, select_l2_entities, [('entities', 'inlist')]),
        (loader, select_l2_indices, [('contrast_indices', 'inlist')]),
        (loader, select_l2_contrasts, [('contrast_info', 'inlist')]),

        (l1_model, l2_model, [('contrast_maps', 'stat_files')]),
        (l1_metadata, l2_model, [('out', 'stat_metadata')]),
        (select_l2_indices, l2_model, [('out', 'contrast_indices')]),
        (select_l2_contrasts, l2_model, [('out', 'contrast_info')]),

        (l2_model, collate_second_level, [('contrast_maps', 'contrast_maps'),
                                          ('contrast_metadata', 'contrast_metadata')]),

        #
        # Plotting
        #
        (l1_model, plot_design, [('design_matrix', 'data')]),

        (loader, get_evs, [('session_info', 'info')]),
        (l1_model, plot_corr, [('design_matrix', 'data')]),
        (get_evs, plot_corr, [('out', 'explanatory_variables')]),

        (l1_model, plot_l1_contrast_matrix, [('contrast_matrix', 'data')]),

        (collate_first_level, plot_l1_contrasts, [('contrast_maps', 'data')]),

        (l2_model, plot_l2_contrast_matrix, [('contrast_matrix', 'data')]),

        (collate_second_level, plot_l2_contrasts, [('contrast_maps', 'data')]),

        #
        # HTML snippets
        #
        (loader, ds_model_warnings, [('warnings', 'in_file')]),
        (select_l1_entities, ds_model_warnings,  [('out', 'entities')]),

        #
        # Derivatives
        #
        (collate_first_level, ds_l1_contrast_maps, [('contrast_maps', 'in_file'),
                                                    ('contrast_metadata', 'entities')]),

        (collate_second_level, ds_l2_contrast_maps, [('contrast_maps', 'in_file'),
                                                     ('contrast_metadata', 'entities')]),

        (select_l1_entities, ds_design, [('out', 'entities')]),
        (plot_design, ds_design, [('figure', 'in_file')]),

        (select_l1_entities, ds_corr, [('out', 'entities')]),
        (plot_corr, ds_corr, [('figure', 'in_file')]),


        (select_l1_entities, ds_l1_contrasts, [('out', 'entities')]),
        (plot_l1_contrast_matrix, ds_l1_contrasts, [('figure', 'in_file')]),

        (collate_first_level, ds_l1_contrast_plots, [('contrast_metadata', 'entities')]),
        (plot_l1_contrasts, ds_l1_contrast_plots, [('figure', 'in_file')]),

        (select_l2_entities, ds_l2_contrasts, [('out', 'entities')]),
        (plot_l2_contrast_matrix, ds_l2_contrasts, [('figure', 'in_file')]),

        (collate_second_level, ds_l2_contrast_plots, [('contrast_metadata', 'entities')]),
        (plot_l2_contrasts, ds_l2_contrast_plots, [('figure', 'in_file')]),
        ])

    return wf



import os
from functools import reduce
import numpy as np
import pandas as pd
import nibabel as nb
from nistats import design_matrix as dm
from nistats import first_level_model as level1
from nistats import second_level_model as level2

from nipype.interfaces.base import (
    LibraryBaseInterface, SimpleInterface, BaseInterfaceInputSpec, TraitedSpec,
    OutputMultiObject, File, traits, isdefined
    )

from ..utils import dict_intersection


class NistatsBaseInterface(LibraryBaseInterface):
    _pkg = 'nistats'


def build_contrast_matrix(contrast_spec, design_matrix,
                          identity=None):
    """Construct contrast matrix and return contrast type

    Parameters
    ----------
    contrast_spec : DataFrame
        Weight matrix with contrasts as rows and regressors as columns
        May have 'type' column indicating T/F test
    design_matrix : DataFrame
        GLM design matrix with regressors as columns and TR time as rows
    identity : list of strings
        Names of explanatory variables to ensure "identity" contrasts are
        provided.

    Returns
    -------
    contrast_matrix : DataFrame
        Weight matrix with contrasts as columns and regressors as rows.
        Regressors match columns (including order) of design matrix.
        Identity contrasts are included.
    contrast_types : Series
        Series of 'T'/'F' indicating the type of test for each column in
        the returned contrast matrix.
        Identity contrasts use T-tests.
    """
    # The basic spec is just a transposed matrix with an optional 'type'
    # column
    # We'll re-transpose and expand this matrix
    init_matrix = contrast_spec.drop('type', axis='columns').T
    init_types = contrast_spec['type'] if 'type' in contrast_spec \
        else pd.Series()

    if identity is None:
        identity = []
    all_cols = init_matrix.columns.tolist()
    all_cols.extend(set(identity) - set(all_cols))

    contrast_matrix = pd.DataFrame(index=design_matrix.columns,
                                   columns=all_cols, data=0)
    contrast_matrix.loc[tuple(init_matrix.axes)] = init_matrix

    contrast_types = pd.Series(index=all_cols, data='T')
    contrast_types[init_types.index] = init_types

    if identity:
        contrast_matrix.loc[identity, identity] = np.eye(len(identity))
        contrast_types[identity] = 'T'

    return contrast_matrix, contrast_types


class FirstLevelModelInputSpec(BaseInterfaceInputSpec):
    bold_file = File(exists=True, mandatory=True)
    mask_file = File(exists=True)
    session_info = traits.Dict()
    contrast_info = File(exists=True)


class FirstLevelModelOutputSpec(TraitedSpec):
    contrast_maps = traits.List(File)
    contrast_metadata = traits.List(traits.Dict)
    design_matrix = File()
    contrast_matrix = File()


class FirstLevelModel(NistatsBaseInterface, SimpleInterface):
    input_spec = FirstLevelModelInputSpec
    output_spec = FirstLevelModelOutputSpec

    def _run_interface(self, runtime):
        info = self.inputs.session_info

        img = nb.load(self.inputs.bold_file)
        vols = img.shape[3]

        events = pd.read_hdf(info['events'], key='events')

        if info['confounds'] is not None and info['confounds'] != 'None':
            confounds = pd.read_hdf(info['confounds'], key='confounds')
            confound_names = confounds.columns.tolist()
            drift_model = None if 'Cosine00' in confound_names else 'cosine'
        else:
            confounds = None
            confound_names = None
            drift_model = 'cosine'

        if isdefined(self.inputs.contrast_info):
            contrast_spec = pd.read_hdf(self.inputs.contrast_info,
                                        key='contrasts')
        else:
            contrast_spec = pd.DataFrame()

        mat = dm.make_design_matrix(
            frame_times=np.arange(vols) * info['repetition_time'],
            paradigm=events.rename(columns={'condition': 'trial_type',
                                            'amplitude': 'modulation'}),
            add_regs=confounds,
            add_reg_names=confound_names,
            drift_model=drift_model,
            )

        # Assume that explanatory variables == HRF-convolved variables
        exp_vars = events['condition'].unique().tolist()

        contrast_matrix, contrast_types = build_contrast_matrix(contrast_spec,
                                                                mat, exp_vars)

        mat.to_csv('design.tsv', sep='\t')
        self._results['design_matrix'] = os.path.join(runtime.cwd,
                                                      'design.tsv')

        contrast_matrix.to_csv('contrasts.tsv', sep='\t')
        self._results['contrast_matrix'] = os.path.join(
            runtime.cwd, 'contrasts.tsv')

        mask_file = self.inputs.mask_file
        if not isdefined(mask_file):
            mask_file = None
        flm = level1.FirstLevelModel(mask=mask_file)
        flm.fit(img, design_matrices=mat)

        contrast_maps = []
        contrast_metadata = []
        stat_fmt = os.path.join(runtime.cwd, '{}.nii.gz').format
        for contrast, ctype in zip(contrast_matrix, contrast_types):
            es = flm.compute_contrast(contrast_matrix[contrast].values,
                                      {'T': 't', 'F': 'F'}[ctype],
                                      output_type='effect_size')
            es_fname = stat_fmt(contrast)
            es.to_filename(es_fname)

            contrast_maps.append(es_fname)
            contrast_metadata.append({'contrast': contrast,
                                      'type': 'effect'})
        self._results['contrast_maps'] = contrast_maps
        self._results['contrast_metadata'] = contrast_metadata

        return runtime


class SecondLevelModelInputSpec(BaseInterfaceInputSpec):
    stat_files = traits.List(traits.List(File(exists=True)), mandatory=True)
    stat_metadata = traits.List(traits.List(traits.Dict))
    contrast_info = File(exists=True)
    contrast_indices = traits.List(traits.Dict)


class SecondLevelModelOutputSpec(TraitedSpec):
    contrast_maps = traits.List(File)
    contrast_metadata = traits.List(traits.Dict)
    contrast_matrix = File()


def _flatten(x):
    return [elem for sublist in x for elem in sublist]


def _match(query, metadata):
    for key, val in query.items():
        if metadata.get(key) != val:
            return False
    return True


class SecondLevelModel(NistatsBaseInterface, SimpleInterface):
    input_spec = SecondLevelModelInputSpec
    output_spec = SecondLevelModelOutputSpec

    def _run_interface(self, runtime):
        model = level2.SecondLevelModel()
        files = []
        # Super inefficient... think more about this later
        for idx in self.inputs.contrast_indices:
            for fname, metadata in zip(_flatten(self.inputs.stat_files),
                                       _flatten(self.inputs.stat_metadata)):
                if _match(idx, metadata):
                    files.append(fname)
                    break
            else:
                raise ValueError

        out_ents = reduce(dict_intersection, self.inputs.contrast_indices)
        in_ents = [{key: val for key, val in index.items() if key not in out_ents}
                   for index in self.inputs.contrast_indices]

        contrast_spec = pd.read_hdf(self.inputs.contrast_info,
                                    key='contrasts')

        contrast_matrix = contrast_spec.drop(columns=['type']).T
        contrast_types = contrast_spec['type']

        contrast_matrix.index = ['_'.join('{}-{}'.format(key, val)
                                          for key, val in ents.items())
                                 for ents in in_ents]
        contrast_matrix.to_csv('contrasts.tsv', sep='\t')
        self._results['contrast_matrix'] = os.path.join(
            runtime.cwd, 'contrasts.tsv')

        out_ents['type'] = 'stat'

        contrast_maps = []
        contrast_metadata = []
        stat_fmt = os.path.join(runtime.cwd, '{}.nii.gz').format
        for contrast, ctype in zip(contrast_matrix, contrast_types):
            intercept = contrast_matrix[contrast]
            data = np.array(files)[intercept != 0].tolist()
            intercept = intercept[intercept != 0]

            model.fit(data, design_matrix=pd.DataFrame({'intercept': intercept}))
            stat_type = {'T': 't', 'F': 'F'}[ctype]

            stat = model.compute_contrast(second_level_stat_type=stat_type)
            stat_fname = stat_fmt(contrast)
            stat.to_filename(stat_fname)

            contrast_maps.append(stat_fname)
            metadata = out_ents.copy()
            metadata['contrast'] = contrast
            contrast_metadata.append(metadata)

        self._results['contrast_maps'] = contrast_maps
        self._results['contrast_metadata'] = contrast_metadata

        return runtime



from matplotlib import pyplot as plt
from .corr import plot_corr_matrix
from .contrasts import plot_contrast_matrix


def plot_and_save(fname, plotter, *args, **kwargs):
    if (kwargs.get('axes'), kwargs.get('ax')) == (None, None):
        fig = plt.figure()
        axes = plt.gca()
        if 'axes' in kwargs:
            kwargs['axes'] = axes
        else:
            kwargs['ax'] = axes
    plotter(*args, **kwargs)
    fig.savefig(fname, bbox_inches='tight')
    plt.close(fig)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
fMRI model-fitting
==================
"""

import sys
import os
import os.path as op
import time
import logging
import warnings
from tempfile import mkdtemp
from argparse import ArgumentParser
from argparse import RawTextHelpFormatter
from multiprocessing import cpu_count

from bids import BIDSLayout, Analysis
from bids.analysis import auto_model

from .. import __version__
from ..workflows import init_fitlins_wf
from ..utils import bids
from ..viz.reports import write_report, parse_directory

logging.addLevelName(25, 'INFO')  # Add a new level between INFO and WARNING
logger = logging.getLogger('cli')
logger.setLevel(25)

INIT_MSG = """
Running FitLins version {version}:
  * Participant list: {subject_list}.
""".format


def _warn_redirect(message, category, filename, lineno, file=None, line=None):
    logger.warning('Captured warning (%s): %s', category, message)


def default_path(arg, base_dir, default_name):
    """Generate absolute path from absolute, relative, or missing path

    Missing paths are given a default value, which may be absolute or relative,
    and relative paths are resolved relative to a base directory.
    """
    path = default_name if arg is None else arg
    if not op.isabs(path):
        path = op.abspath(op.join(base_dir, path))
    return path


def get_parser():
    """Build parser object"""
    verstr = 'fitlins v{}'.format(__version__)

    parser = ArgumentParser(description='FitLins: Workflows for Fitting Linear models to fMRI',
                            formatter_class=RawTextHelpFormatter)

    # Arguments as specified by BIDS-Apps
    # required, positional arguments
    # IMPORTANT: they must go directly with the parser object
    parser.add_argument('bids_dir', action='store', type=op.abspath,
                        help='the root folder of a BIDS valid dataset (sub-XXXXX folders should '
                             'be found at the top level in this folder).')
    parser.add_argument('output_dir', action='store', type=op.abspath,
                        help='the output path for the outcomes of preprocessing and visual '
                             'reports')
    parser.add_argument('analysis_level', choices=['run', 'session', 'participant', 'dataset'],
                        help='processing stage to be runa (see BIDS-Apps specification).')

    # optional arguments
    parser.add_argument('-v', '--version', action='version', version=verstr)

    g_bids = parser.add_argument_group('Options for filtering BIDS queries')
    g_bids.add_argument('--participant-label', action='store', nargs='+', default=None,
                        help='one or more participant identifiers (the sub- prefix can be '
                             'removed)')
    g_bids.add_argument('-m', '--model', action='store', default='model.json',
                        help='location of BIDS model description (default bids_dir/model.json)')
    g_bids.add_argument('-p', '--preproc-dir', action='store', default='fmriprep',
                        help='location of preprocessed data (if relative path, search '
                             'bids_dir/derivatives, followed by output_dir)')
    g_bids.add_argument('--derivative-label', action='store', type=str,
                        help='execution label to append to derivative directory name')
    g_bids.add_argument('--space', action='store',
                        choices=['MNI152NLin2009cAsym'], default='MNI152NLin2009cAsym',
                        help='registered space of input datasets')
    g_bids.add_argument('--include', action='store', default=None,
                        help='regex pattern to include files')
    g_bids.add_argument('--exclude', action='store', default=None,
                        help='regex pattern to exclude files')

    g_perfm = parser.add_argument_group('Options to handle performance')
    g_perfm.add_argument('--n-cpus', action='store', default=0, type=int,
                         help='maximum number of threads across all processes')
    g_perfm.add_argument('--debug', action='store_true', default=False,
                         help='run debug version of workflow')

    g_other = parser.add_argument_group('Other options')
    g_other.add_argument('-w', '--work-dir', action='store', type=op.abspath,
                         help='path where intermediate results should be stored')

    return parser


def run_fitlins(argv=None):
    warnings.showwarning = _warn_redirect
    opts = get_parser().parse_args(argv)
    if opts.debug:
        logger.setLevel(logging.DEBUG)

    subject_list = None
    if opts.participant_label is not None:
        subject_list = bids.collect_participants(
            opts.bids_dir, participant_label=opts.participant_label)

    ncpus = opts.n_cpus
    if ncpus < 1:
        ncpus = cpu_count()

    plugin_settings = {
        'plugin': 'MultiProc',
        'plugin_args': {
            'n_procs': ncpus,
            'raise_insufficient': False,
            'maxtasksperchild': 1,
        }
    }

    # Build main workflow
    logger.log(25, INIT_MSG(
        version=__version__,
        subject_list=subject_list)
    )

    model = default_path(opts.model, opts.bids_dir, 'model.json')
    if opts.model in (None, 'default') and not op.exists(model):
        model = 'default'

    preproc_dir = default_path(opts.preproc_dir,
                               op.join(opts.bids_dir, 'derivatives'),
                               'fmriprep')
    if not op.exists(preproc_dir):
        preproc_dir = default_path(opts.preproc_dir, opts.output_dir, 'fmriprep')
        if not op.exists(preproc_dir):
            raise RuntimeError("Preprocessed data could not be found")

    pipeline_name = 'fitlins'
    if opts.derivative_label:
        pipeline_name += '_' + opts.derivative_label
    deriv_dir = op.join(opts.output_dir, pipeline_name)
    os.makedirs(deriv_dir, exist_ok=True)

    bids.write_derivative_description(opts.bids_dir, deriv_dir)

    work_dir = mkdtemp() if opts.work_dir is None else opts.work_dir

    fitlins_wf = init_fitlins_wf(
        opts.bids_dir, preproc_dir, deriv_dir, opts.space, model=model,
        participants=subject_list, base_dir=work_dir,
        include_pattern=opts.include, exclude_pattern=opts.exclude
        )

    retcode = 0
    try:
        fitlins_wf.run(**plugin_settings)
    except Exception:
        retcode = 1

    layout = BIDSLayout(opts.bids_dir)
    models = auto_model(layout) if model == 'default' else [model]

    run_context = {'version': __version__,
                   'command': ' '.join(sys.argv),
                   'timestamp': time.strftime('%Y-%m-%d %H:%M:%S %z'),
                   }

    for model in models:
        analysis = Analysis(layout, model=model)
        report_dicts = parse_directory(deriv_dir, work_dir, analysis)
        write_report('unknown', report_dicts, run_context, deriv_dir)

    return retcode


def main():
    sys.exit(run_fitlins(sys.argv[1:]))


if __name__ == '__main__':
    raise RuntimeError("fitlins/cli/run.py should not be run directly;\n"
                       "Please `pip install` fitlins and use the `fitlins` command")



from nipype.interfaces.io import IOBase, add_traits
from nipype.interfaces.base import DynamicTraitedSpec


class MergeAll(IOBase):
    input_spec = DynamicTraitedSpec
    output_spec = DynamicTraitedSpec

    def __init__(self, fields=None):
        super(MergeAll, self).__init__()
        if not fields:
            raise ValueError("Fields must be a non-empty list")

        self._fields = fields
        add_traits(self.inputs, fields)

    def _add_output_traits(self, base):
        return add_traits(base, self._fields)

    def _list_outputs(self):
        outputs = self._outputs().get()
        lengths = None
        for key in self._fields:
            val = getattr(self.inputs, key)
            _lengths = list(map(len, val))
            if lengths is None:
                lengths = _lengths
            elif _lengths != lengths:
                raise ValueError("List lengths must be consistent across fields")
            outputs[key] = [elem for sublist in val for elem in sublist]

        return outputs



import os
from functools import reduce
from pathlib import Path
from gzip import GzipFile
import json
import shutil
import numpy as np
import nibabel as nb

from collections import defaultdict

from nipype import logging
from nipype.utils.filemanip import makedirs, copyfile
from nipype.interfaces.base import (
    BaseInterfaceInputSpec, TraitedSpec, SimpleInterface,
    InputMultiPath, OutputMultiPath, File, Directory,
    traits, isdefined
    )
from nipype.interfaces.io import IOBase

from ..utils import dict_intersection, snake_to_camel

iflogger = logging.getLogger('nipype.interface')


def bids_split_filename(fname):
    """Split a filename into parts: path, base filename, and extension

    Respects multi-part file types used in BIDS standard and draft extensions

    Largely copied from nipype.utils.filemanip.split_filename

    Parameters
    ----------
    fname : str
        file or path name

    Returns
    -------
    pth : str
        path of fname
    fname : str
        basename of filename, without extension
    ext : str
        file extension of fname
    """
    special_extensions = [
        ".R.surf.gii", ".L.surf.gii",
        ".R.func.gii", ".L.func.gii",
        ".nii.gz", ".tsv.gz",
        ]

    pth = os.path.dirname(fname)
    fname = os.path.basename(fname)

    for special_ext in special_extensions:
        if fname.lower().endswith(special_ext.lower()):
            ext_len = len(special_ext)
            ext = fname[-ext_len:]
            fname = fname[:-ext_len]
            break
    else:
        fname, ext = os.path.splitext(fname)

    return pth, fname, ext


def _ensure_model(model):
    model = getattr(model, 'filename', model)

    if isinstance(model, str):
        if os.path.exists(model):
            with open(model) as fobj:
                model = json.load(fobj)
        else:
            model = json.loads(model)
    return model


class ModelSpecLoaderInputSpec(BaseInterfaceInputSpec):
    bids_dir = Directory(exists=True,
                         mandatory=True,
                         desc='BIDS dataset root directory')
    model = traits.Either('default', InputMultiPath(File(exists=True)),
                          desc='Model filename')
    selectors = traits.Dict(desc='Limit models to those with matching inputs')


class ModelSpecLoaderOutputSpec(TraitedSpec):
    model_spec = OutputMultiPath(traits.Dict())


class ModelSpecLoader(SimpleInterface):
    input_spec = ModelSpecLoaderInputSpec
    output_spec = ModelSpecLoaderOutputSpec

    def _run_interface(self, runtime):
        import bids
        from bids.analysis import auto_model
        models = self.inputs.model
        if not isinstance(models, list):
            layout = bids.BIDSLayout(self.inputs.bids_dir)

            if not isdefined(models):
                models = layout.get(type='model')
                if not models:
                    raise ValueError("No models found")
            elif models == 'default':
                models = auto_model(layout)

        models = [_ensure_model(m) for m in models]

        if self.inputs.selectors:
            # This is almost certainly incorrect
            models = [model for model in models
                      if all(val in model['input'].get(key, [val])
                             for key, val in self.inputs.selectors.items())]

        self._results['model_spec'] = models

        return runtime


IMPUTATION_SNIPPET = """\
<div class="warning">
    The following confounds had NaN values for the first volume: {}.
    The mean of non-zero values for the remaining entries was imputed.
    If another strategy is desired, it must be explicitly specified in
    the model.
</div>
"""


class LoadBIDSModelInputSpec(BaseInterfaceInputSpec):
    bids_dir = Directory(exists=True,
                         mandatory=True,
                         desc='BIDS dataset root directory')
    preproc_dir = Directory(exists=True,
                            desc='Optional preprocessed files directory')
    model = traits.Dict(desc='Model specification', mandatory=True)
    selectors = traits.Dict(desc='Limit collected sessions', usedefault=True)
    include_pattern = InputMultiPath(
        traits.Str, xor=['exclude_pattern'],
        desc='Patterns to select sub-directories of BIDS root')
    exclude_pattern = InputMultiPath(
        traits.Str, xor=['include_pattern'],
        desc='Patterns to ignore sub-directories of BIDS root')


class LoadBIDSModelOutputSpec(TraitedSpec):
    session_info = traits.List(traits.Dict())
    contrast_info = traits.List(traits.List(File()))
    contrast_indices = traits.List(traits.List(traits.List(traits.Dict)))
    entities = traits.List(traits.List(traits.Dict()))
    warnings = traits.List(File)


class LoadBIDSModel(SimpleInterface):
    input_spec = LoadBIDSModelInputSpec
    output_spec = LoadBIDSModelOutputSpec

    def _run_interface(self, runtime):
        import bids
        bids.config.set_options(loop_preproc=True)
        include = self.inputs.include_pattern
        exclude = self.inputs.exclude_pattern
        if not isdefined(include):
            include = None
        if not isdefined(exclude):
            exclude = None

        paths = [(self.inputs.bids_dir, 'bids')]
        if isdefined(self.inputs.preproc_dir):
            paths.append((self.inputs.preproc_dir, ['bids', 'derivatives']))
        layout = bids.BIDSLayout(paths, include=include, exclude=exclude)

        selectors = self.inputs.selectors

        analysis = bids.Analysis(model=self.inputs.model, layout=layout)
        analysis.setup(drop_na=False, **selectors)
        self._load_level1(runtime, analysis)
        self._load_higher_level(runtime, analysis)

        # Debug - remove, eventually
        runtime.analysis = analysis

        return runtime

    def _load_level1(self, runtime, analysis):
        block = analysis.blocks[0]
        block_subdir = Path(runtime.cwd) / block.level
        block_subdir.mkdir(parents=True, exist_ok=True)

        entities = []
        session_info = []
        contrast_indices = []
        contrast_info = []
        warnings = []
        for paradigm, _, ents in block.get_design_matrix(
                block.model['HRF_variables'], mode='sparse', force=True):
            info = {}

            space = analysis.layout.get_spaces(type='preproc',
                                               extensions=['.nii', '.nii.gz'])[0]
            preproc_files = analysis.layout.get(type='preproc',
                                                extensions=['.nii', '.nii.gz'],
                                                space=space,
                                                **ents)
            if len(preproc_files) != 1:
                raise ValueError('Too many BOLD files found')

            fname = preproc_files[0].filename

            # Required field in seconds
            TR = analysis.layout.get_metadata(fname, type='bold',
                                              full_search=True)['RepetitionTime']
            dense_vars = set(block.model['variables']) - set(block.model['HRF_variables'])

            _, confounds, _ = block.get_design_matrix(dense_vars,
                                                      mode='dense',
                                                      force=True,
                                                      sampling_rate=1/TR,
                                                      **ents)[0]

            ent_string = '_'.join('{}-{}'.format(key, val)
                                  for key, val in ents.items())

            events_file = block_subdir / '{}_events.h5'.format(ent_string)
            paradigm.to_hdf(events_file, key='events')

            imputed = []
            if confounds is not None:
                # Note that FMRIPREP includes CosineXX columns to accompany
                # t/aCompCor
                # We may want to add criteria to include HPF columns that are not
                # explicitly listed in the model
                names = [col for col in confounds.columns
                         if col.startswith('NonSteadyStateOutlier') or
                         col in block.model['variables']]
                confounds = confounds[names]

                # These confounds are defined pairwise with the current volume
                # and its predecessor, and thus may be undefined (have value
                # NaN) at the first volume.
                # In these cases, we impute the mean non-zero value, for the
                # expected NaN only.
                # Any other NaNs must be handled by an explicit transform in
                # the BIDS model.
                for imputable in ('FramewiseDisplacement',
                                  'stdDVARS', 'non-stdDVARS',
                                  'vx-wisestdDVARS'):
                    if imputable in confounds.columns:
                        vals = confounds[imputable].values
                        if not np.isnan(vals[0]):
                            continue

                        # Impute the mean non-zero, non-NaN value
                        confounds[imputable][0] = np.nanmean(vals[vals != 0])
                        imputed.append(imputable)

                if np.isnan(confounds.values).any():
                    iflogger.warning('Unexpected NaNs found in confounds; '
                                     'regression may fail.')

                confounds_file = block_subdir / '{}_confounds.h5'.format(ent_string)
                confounds.to_hdf(confounds_file, key='confounds')

            else:
                confounds_file = None

            info['events'] = str(events_file)
            info['confounds'] = str(confounds_file)
            info['repetition_time'] = TR

            # Transpose so each contrast gets a row of data instead of column
            contrasts, index, _ = block.get_contrasts(**ents)[0]

            contrast_type_map = defaultdict(lambda: 'T')
            contrast_type_map.update({contrast['name']: contrast['type']
                                      for contrast in block.contrasts})
            contrast_type_list = [contrast_type_map[contrast]
                                  for contrast in contrasts.columns]

            contrasts = contrasts.T
            # Add test indicator column
            contrasts['type'] = contrast_type_list

            contrasts_file = block_subdir / '{}_contrasts.h5'.format(ent_string)
            contrasts_file.parent.mkdir(parents=True, exist_ok=True)
            contrasts.to_hdf(contrasts_file, key='contrasts')

            warning_file = block_subdir / '{}_warning.html'.format(ent_string)
            with warning_file.open('w') as fobj:
                if imputed:
                    fobj.write(IMPUTATION_SNIPPET.format(', '.join(imputed)))

            entities.append(ents)
            session_info.append(info)
            contrast_indices.append(index.to_dict('records'))
            contrast_info.append(str(contrasts_file))
            warnings.append(str(warning_file))

        self._results['session_info'] = session_info
        self._results['warnings'] = warnings
        self._results.setdefault('entities', []).append(entities)
        self._results.setdefault('contrast_indices', []).append(contrast_indices)
        self._results.setdefault('contrast_info', []).append(contrast_info)

    def _load_higher_level(self, runtime, analysis):
        cwd = Path(runtime.cwd)
        for block in analysis.blocks[1:]:
            block_subdir = cwd / block.level
            block_subdir.mkdir(parents=True, exist_ok=True)

            entities = []
            contrast_indices = []
            contrast_info = []
            for contrasts, index, ents in block.get_contrasts():
                if contrasts.empty:
                    continue

                # The contrast index is the name of the input contrasts,
                # which will very frequently be non-unique
                # Hence, add the contrast to the index (table of entities)
                # and switch to a matching numeric index
                index['contrast'] = contrasts.index
                contrasts.index = index.index

                contrast_type_map = defaultdict(lambda: 'T')
                contrast_type_map.update({contrast['name']: contrast['type']
                                          for contrast in block.contrasts})
                contrast_type_list = [contrast_type_map[contrast]
                                      for contrast in contrasts.columns]

                indices = index.to_dict('records')

                # Entities for a given contrast matrix include the intersection of
                # entities of inputs, e.g., if this level is within-subject, the
                # subject should persist
                out_ents = reduce(dict_intersection, indices)
                # Explicit entities take precedence over derived
                out_ents.update(ents)
                # Input-level contrasts will be overridden by the current level
                out_ents.pop('contrast', None)

                ent_string = '_'.join('{}-{}'.format(key, val)
                                      for key, val in out_ents.items())

                # Transpose so each contrast gets a row of data instead of column
                contrasts = contrasts.T
                # Add test indicator column
                contrasts['type'] = contrast_type_list

                contrasts_file = block_subdir / '{}_contrasts.h5'.format(ent_string)
                contrasts_file.parent.mkdir(parents=True, exist_ok=True)
                contrasts.to_hdf(contrasts_file, key='contrasts')

                entities.append(out_ents)
                contrast_indices.append(indices)
                contrast_info.append(str(contrasts_file))

            self._results['entities'].append(entities)
            self._results['contrast_info'].append(contrast_info)
            self._results['contrast_indices'].append(contrast_indices)


class BIDSSelectInputSpec(BaseInterfaceInputSpec):
    bids_dir = Directory(exists=True,
                         mandatory=True,
                         desc='BIDS dataset root directories')
    preproc_dir = Directory(exists=True,
                            desc='Optional preprocessed files directory')
    entities = InputMultiPath(traits.Dict(), mandatory=True)
    selectors = traits.Dict(desc='Additional selectors to be applied',
                            usedefault=True)


class BIDSSelectOutputSpec(TraitedSpec):
    bold_files = OutputMultiPath(File)
    mask_files = OutputMultiPath(traits.Either(File, None))
    entities = OutputMultiPath(traits.Dict)


class BIDSSelect(SimpleInterface):
    input_spec = BIDSSelectInputSpec
    output_spec = BIDSSelectOutputSpec

    def _run_interface(self, runtime):
        import bids
        paths = [(self.inputs.bids_dir, 'bids')]
        if isdefined(self.inputs.preproc_dir):
            paths.append((self.inputs.preproc_dir, ['bids', 'derivatives']))
        layout = bids.BIDSLayout(paths)

        bold_files = []
        mask_files = []
        entities = []
        for ents in self.inputs.entities:
            selectors = {**self.inputs.selectors, **ents}
            bold_file = layout.get(extensions=['.nii', '.nii.gz'], **selectors)

            if len(bold_file) == 0:
                raise FileNotFoundError(
                    "Could not find BOLD file in {} with entities {}"
                    "".format(self.inputs.bids_dir, selectors))
            elif len(bold_file) > 1:
                raise ValueError(
                    "Non-unique BOLD file in {} with entities {}.\n"
                    "Matches:\n\t{}"
                    "".format(self.inputs.bids_dir, selectors,
                              "\n\t".join(
                                  '{} ({})'.format(
                                      f.filename,
                                      layout.files[f.filename].entities)
                                  for f in bold_file)))

            # Select exactly matching mask file (may be over-cautious)
            bold_ents = layout.parse_file_entities(
                bold_file[0].filename)
            bold_ents['type'] = 'brainmask'
            mask_file = layout.get(extensions=['.nii', '.nii.gz'], **bold_ents)
            bold_ents.pop('type')

            bold_files.append(bold_file[0].filename)
            mask_files.append(mask_file[0].filename if mask_file else None)
            entities.append(bold_ents)

        self._results['bold_files'] = bold_files
        self._results['mask_files'] = mask_files
        self._results['entities'] = entities

        return runtime


def _copy_or_convert(in_file, out_file):
    in_ext = bids_split_filename(in_file)[2]
    out_ext = bids_split_filename(out_file)[2]

    # Copy if filename matches
    if in_ext == out_ext:
        copyfile(in_file, out_file, copy=True, use_hardlink=True)
        return

    # gzip/gunzip if it's easy
    if in_ext == out_ext + '.gz' or in_ext + '.gz' == out_ext:
        read_open = GzipFile if in_ext.endswith('.gz') else open
        write_open = GzipFile if out_ext.endswith('.gz') else open
        with read_open(in_file, mode='rb') as in_fobj:
            with write_open(out_file, mode='wb') as out_fobj:
                shutil.copyfileobj(in_fobj, out_fobj)
        return

    # Let nibabel take a shot
    try:
        nb.save(nb.load(in_file), out_file)
    except Exception:
        pass
    else:
        return

    raise RuntimeError("Cannot convert {} to {}".format(in_ext, out_ext))


class BIDSDataSinkInputSpec(BaseInterfaceInputSpec):
    base_directory = Directory(
        mandatory=True,
        desc='Path to BIDS (or derivatives) root directory')
    in_file = InputMultiPath(File(exists=True), mandatory=True)
    entities = InputMultiPath(traits.Dict, usedefault=True,
                              desc='Per-file entities to include in filename')
    fixed_entities = traits.Dict(usedefault=True,
                                 desc='Entities to include in all filenames')
    path_patterns = InputMultiPath(
        traits.Str, desc='BIDS path patterns describing format of file names')


class BIDSDataSinkOutputSpec(TraitedSpec):
    out_file = OutputMultiPath(File, desc='output file')


class BIDSDataSink(IOBase):
    input_spec = BIDSDataSinkInputSpec
    output_spec = BIDSDataSinkOutputSpec

    _always_run = True

    def _list_outputs(self):
        import bids
        base_dir = self.inputs.base_directory

        os.makedirs(base_dir, exist_ok=True)

        layout = bids.BIDSLayout(base_dir)
        path_patterns = self.inputs.path_patterns
        if not isdefined(path_patterns):
            path_patterns = None

        out_files = []
        for entities, in_file in zip(self.inputs.entities,
                                     self.inputs.in_file):
            ents = {**self.inputs.fixed_entities}
            ents.update(entities)

            ents = {k: snake_to_camel(str(v)) for k, v in ents.items()}

            out_fname = os.path.join(
                base_dir, layout.build_path(ents, path_patterns))
            makedirs(os.path.dirname(out_fname), exist_ok=True)

            _copy_or_convert(in_file, out_fname)
            out_files.append(out_fname)

        return {'out_file': out_files}
